from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Optional, Union

from acryl_datahub_cloud.sdk.assertion.assertion_base import (
    AssertionMode,
)
from acryl_datahub_cloud.sdk.assertion.smart_column_metric_assertion import (
    SmartColumnMetricAssertion,
)
from acryl_datahub_cloud.sdk.assertion_client.helpers import (
    DEFAULT_CREATED_BY,
    _merge_field,
    _print_experimental_warning,
    _validate_required_field,
    retrieve_assertion_and_monitor_by_urn,
)
from acryl_datahub_cloud.sdk.assertion_input.assertion_input import (
    AssertionIncidentBehaviorInputTypes,
    DetectionMechanismInputTypes,
    ExclusionWindowInputTypes,
    InferenceSensitivity,
)
from acryl_datahub_cloud.sdk.assertion_input.column_metric_constants import (
    MetricInputType,
)
from acryl_datahub_cloud.sdk.assertion_input.smart_column_metric_assertion_input import (
    _SmartColumnMetricAssertionInput,
)
from acryl_datahub_cloud.sdk.entities.assertion import Assertion, TagsInputType
from acryl_datahub_cloud.sdk.entities.monitor import Monitor
from acryl_datahub_cloud.sdk.errors import SDKUsageError
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, CorpUserUrn, DatasetUrn, MonitorUrn

if TYPE_CHECKING:
    from datahub.sdk.main_client import DataHubClient

logger = logging.getLogger(__name__)


class SmartColumnMetricAssertionClient:
    """Client for managing smart column metric assertions."""

    def __init__(self, client: "DataHubClient"):
        self.client = client

    def _validate_required_smart_column_fields_for_creation(
        self,
        column_name: Optional[str],
        metric_type: Optional[MetricInputType],
    ) -> None:
        """Validate required fields for smart column metric assertion creation."""
        _validate_required_field(
            column_name, "column_name", "when creating a new assertion (urn is None)"
        )
        _validate_required_field(
            metric_type, "metric_type", "when creating a new assertion (urn is None)"
        )

    def _validate_required_smart_column_fields_for_update(
        self,
        column_name: Optional[str],
        metric_type: Optional[MetricInputType],
        assertion_urn: Union[str, AssertionUrn],
    ) -> None:
        """Validate required fields after attempting to fetch from existing assertion."""
        context = f"and not found in existing assertion {assertion_urn}. The existing assertion may be invalid or corrupted."
        _validate_required_field(column_name, "column_name", context)
        _validate_required_field(metric_type, "metric_type", context)

    def _retrieve_assertion_and_monitor(
        self,
        urn: Union[str, AssertionUrn],
        dataset_urn: Union[str, DatasetUrn],
    ) -> tuple[Optional[Assertion], MonitorUrn, Optional[Monitor]]:
        """Retrieve the assertion and monitor entities from the DataHub instance.

        Args:
            urn: The assertion URN.
            dataset_urn: The dataset URN.

        Returns:
            The assertion and monitor entities.
        """
        return retrieve_assertion_and_monitor_by_urn(self.client, urn, dataset_urn)

    def sync_smart_column_metric_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        column_name: Optional[str] = None,
        metric_type: Optional[MetricInputType] = None,
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        sensitivity: Optional[Union[str, InferenceSensitivity]] = None,
        exclusion_windows: Optional[ExclusionWindowInputTypes] = None,
        training_data_lookback_days: Optional[int] = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
    ) -> SmartColumnMetricAssertion:
        _print_experimental_warning()
        now_utc = datetime.now(timezone.utc)

        if updated_by is None:
            logger.warning(
                f"updated_by is not set, using {DEFAULT_CREATED_BY} as a placeholder"
            )
            updated_by = DEFAULT_CREATED_BY

        # 1. Retrieve and merge the assertion input with any existing assertion and monitor entities,
        # or build a new assertion input if the assertion does not exist:
        assertion_input = (
            self._retrieve_and_merge_smart_column_metric_assertion_and_monitor(
                dataset_urn=dataset_urn,
                urn=urn,
                column_name=column_name,
                metric_type=metric_type,
                display_name=display_name,
                enabled=enabled,
                detection_mechanism=detection_mechanism,
                sensitivity=sensitivity,
                exclusion_windows=exclusion_windows,
                training_data_lookback_days=training_data_lookback_days,
                incident_behavior=incident_behavior,
                tags=tags,
                updated_by=updated_by,
                now_utc=now_utc,
                schedule=schedule,
            )
        )

        # 2. Upsert the assertion and monitor entities:
        assertion_entity, monitor_entity = (
            assertion_input.to_assertion_and_monitor_entities()
        )
        # If assertion upsert fails, we won't try to upsert the monitor
        self.client.entities.upsert(assertion_entity)
        # TODO: Wrap monitor upsert in a try-except and delete the assertion if monitor upsert fails (once delete is implemented https://linear.app/acryl-data/issue/OBS-1350/add-delete-method-to-entity-clientpy)
        # try:
        self.client.entities.upsert(monitor_entity)
        # except Exception as e:
        #     logger.error(f"Error upserting monitor: {e}")
        #     self.client.entities.delete(assertion_entity)
        #     raise e

        return SmartColumnMetricAssertion._from_entities(
            assertion_entity, monitor_entity
        )

    def _extract_fields_from_existing_assertion(
        self,
        maybe_assertion_entity: Optional[Assertion],
        column_name: Optional[str],
        metric_type: Optional[MetricInputType],
    ) -> tuple[Optional[str], Optional[MetricInputType]]:
        """Extract missing fields from existing assertion entity."""
        if maybe_assertion_entity is None:
            return column_name, metric_type

        assertion_info = maybe_assertion_entity.info
        if not (
            hasattr(assertion_info, "fieldMetricAssertion")
            and assertion_info.fieldMetricAssertion
        ):
            return column_name, metric_type

        field_metric_assertion = assertion_info.fieldMetricAssertion

        # Extract column_name
        if (
            column_name is None
            and hasattr(field_metric_assertion, "field")
            and hasattr(field_metric_assertion.field, "path")
        ):
            column_name = field_metric_assertion.field.path

        # Extract metric_type
        if metric_type is None and hasattr(field_metric_assertion, "metric"):
            metric_type = field_metric_assertion.metric

        return column_name, metric_type

    def _create_existing_assertion_from_entities(
        self,
        maybe_assertion_entity: Optional[Assertion],
        maybe_monitor_entity: Optional[Monitor],
        monitor_urn: MonitorUrn,
        enabled: Optional[bool],
    ) -> Optional[SmartColumnMetricAssertion]:
        """Create existing assertion object from entities, handling missing monitor case."""
        if maybe_assertion_entity and maybe_monitor_entity:
            return SmartColumnMetricAssertion._from_entities(
                maybe_assertion_entity, maybe_monitor_entity
            )

        if maybe_assertion_entity and not maybe_monitor_entity:
            monitor_mode = (
                "ACTIVE" if enabled else "INACTIVE" if enabled is not None else "ACTIVE"
            )
            return SmartColumnMetricAssertion._from_entities(
                maybe_assertion_entity,
                Monitor(id=monitor_urn, info=("ASSERTION", monitor_mode)),
            )

        return None

    def _retrieve_and_merge_smart_column_metric_assertion_and_monitor(
        self,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]],
        column_name: Optional[str],
        metric_type: Optional[MetricInputType],
        display_name: Optional[str],
        enabled: Optional[bool],
        detection_mechanism: DetectionMechanismInputTypes,
        sensitivity: Optional[Union[str, InferenceSensitivity]],
        exclusion_windows: Optional[ExclusionWindowInputTypes],
        training_data_lookback_days: Optional[int],
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes],
        tags: Optional[TagsInputType],
        updated_by: Union[str, CorpUserUrn],
        now_utc: datetime,
        schedule: Optional[Union[str, models.CronScheduleClass]],
    ) -> _SmartColumnMetricAssertionInput:
        # 1. If urn is not provided, validate required fields and build a new assertion input directly
        if urn is None:
            logger.info("URN is not set, building a new assertion input")
            self._validate_required_smart_column_fields_for_creation(
                column_name, metric_type
            )
            assert column_name is not None and metric_type is not None, (
                "Fields guaranteed non-None after validation"
            )
            return _SmartColumnMetricAssertionInput(
                urn=None,
                entity_client=self.client.entities,
                dataset_urn=dataset_urn,
                column_name=column_name,
                metric_type=metric_type,
                display_name=display_name,
                enabled=enabled if enabled is not None else True,
                detection_mechanism=detection_mechanism,
                sensitivity=sensitivity,
                exclusion_windows=exclusion_windows,
                training_data_lookback_days=training_data_lookback_days,
                incident_behavior=incident_behavior,
                tags=tags,
                created_by=updated_by,
                created_at=now_utc,
                updated_by=updated_by,
                updated_at=now_utc,
                schedule=schedule,
            )

        # 2. Retrieve any existing assertion and monitor entities
        maybe_assertion_entity, monitor_urn, maybe_monitor_entity = (
            self._retrieve_assertion_and_monitor(urn, dataset_urn)
        )

        # 3. Extract missing required fields from existing assertion
        column_name, metric_type = self._extract_fields_from_existing_assertion(
            maybe_assertion_entity, column_name, metric_type
        )

        # 4. Validate required fields
        if maybe_assertion_entity is None:
            self._validate_required_smart_column_fields_for_creation(
                column_name, metric_type
            )
        else:
            self._validate_required_smart_column_fields_for_update(
                column_name, metric_type, urn
            )

        assert column_name is not None and metric_type is not None, (
            "Fields guaranteed non-None after validation"
        )

        # 5. If the assertion does not exist, build and return a new assertion input
        if not maybe_assertion_entity:
            logger.info(
                f"No existing assertion entity found for assertion urn {urn}, building a new assertion input"
            )
            return _SmartColumnMetricAssertionInput(
                urn=urn,
                entity_client=self.client.entities,
                dataset_urn=dataset_urn,
                column_name=column_name,
                metric_type=metric_type,
                display_name=display_name,
                enabled=enabled if enabled is not None else True,
                detection_mechanism=detection_mechanism,
                sensitivity=sensitivity,
                exclusion_windows=exclusion_windows,
                training_data_lookback_days=training_data_lookback_days,
                incident_behavior=incident_behavior,
                tags=tags,
                created_by=updated_by,
                created_at=now_utc,
                updated_by=updated_by,
                updated_at=now_utc,
                schedule=schedule,
            )

        # 6. Build assertion input for validation
        assertion_input = _SmartColumnMetricAssertionInput(
            urn=urn,
            entity_client=self.client.entities,
            dataset_urn=dataset_urn,
            column_name=column_name,
            metric_type=metric_type,
            display_name=display_name,
            detection_mechanism=detection_mechanism,
            sensitivity=sensitivity,
            exclusion_windows=exclusion_windows,
            training_data_lookback_days=training_data_lookback_days,
            incident_behavior=incident_behavior,
            tags=tags,
            created_by=updated_by,
            created_at=now_utc,
            updated_by=updated_by,
            updated_at=now_utc,
            schedule=schedule,
        )

        # 7. Create existing assertion from entities
        existing_assertion = self._create_existing_assertion_from_entities(
            maybe_assertion_entity, maybe_monitor_entity, monitor_urn, enabled
        )
        assert existing_assertion is not None, (
            "existing_assertion is guaranteed non-None after early return check"
        )

        # 8. Check for any issues e.g. different dataset urns
        if (
            existing_assertion
            and hasattr(existing_assertion, "dataset_urn")
            and existing_assertion.dataset_urn != assertion_input.dataset_urn
        ):
            raise SDKUsageError(
                f"Dataset URN mismatch, existing assertion: {existing_assertion.dataset_urn} != new assertion: {dataset_urn}"
            )

        # 9. Merge the existing assertion with the validated input
        merged_assertion_input = self._merge_smart_column_metric_input(
            dataset_urn=dataset_urn,
            column_name=column_name,
            metric_type=metric_type,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            schedule=schedule,
            detection_mechanism=detection_mechanism,
            sensitivity=sensitivity,
            exclusion_windows=exclusion_windows,
            training_data_lookback_days=training_data_lookback_days,
            incident_behavior=incident_behavior,
            tags=tags,
            now_utc=now_utc,
            assertion_input=assertion_input,
            maybe_assertion_entity=maybe_assertion_entity,
            maybe_monitor_entity=maybe_monitor_entity,
            existing_assertion=existing_assertion,
        )

        return merged_assertion_input

    def _merge_smart_column_metric_input(
        self,
        dataset_urn: Union[str, DatasetUrn],
        column_name: str,
        metric_type: MetricInputType,
        urn: Union[str, AssertionUrn],
        display_name: Optional[str],
        enabled: Optional[bool],
        detection_mechanism: DetectionMechanismInputTypes,
        sensitivity: Optional[Union[str, InferenceSensitivity]],
        exclusion_windows: Optional[ExclusionWindowInputTypes],
        training_data_lookback_days: Optional[int],
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes],
        tags: Optional[TagsInputType],
        schedule: Optional[Union[str, models.CronScheduleClass]],
        now_utc: datetime,
        assertion_input: _SmartColumnMetricAssertionInput,
        maybe_assertion_entity: Optional[Assertion],
        maybe_monitor_entity: Optional[Monitor],
        existing_assertion: SmartColumnMetricAssertion,
    ) -> _SmartColumnMetricAssertionInput:
        """Merge the input with the existing assertion and monitor entities.

        Args:
            dataset_urn: The urn of the dataset to be monitored.
            column_name: The name of the column to be monitored.
            metric_type: The type of the metric to be monitored.
            urn: The urn of the assertion.
            display_name: The display name of the assertion.
            enabled: Whether the assertion is enabled.
            detection_mechanism: The detection mechanism to be used for the assertion.
            sensitivity: The sensitivity to be applied to the assertion.
            exclusion_windows: The exclusion windows to be applied to the assertion.
            training_data_lookback_days: The training data lookback days to be applied to the assertion.
            incident_behavior: The incident behavior to be applied to the assertion.
            tags: The tags to be applied to the assertion.
            schedule: The schedule to be applied to the assertion.
            now_utc: The current UTC time from when the function is called.
            assertion_input: The validated input to the function.
            maybe_assertion_entity: The existing assertion entity from the DataHub instance.
            maybe_monitor_entity: The existing monitor entity from the DataHub instance.
            existing_assertion: The existing assertion from the DataHub instance.

        Returns:
            The merged assertion input.
        """
        merged_assertion_input = _SmartColumnMetricAssertionInput(
            urn=urn,
            entity_client=self.client.entities,
            dataset_urn=dataset_urn,
            column_name=_merge_field(
                input_field_value=column_name,
                input_field_name="column_name",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=SmartColumnMetricAssertion._get_column_name(
                    maybe_assertion_entity
                )
                if maybe_assertion_entity
                else None,
            ),
            metric_type=_merge_field(
                input_field_value=metric_type,
                input_field_name="metric_type",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=SmartColumnMetricAssertion._get_metric_type(
                    maybe_assertion_entity
                )
                if maybe_assertion_entity
                else None,
            ),
            display_name=_merge_field(
                input_field_value=display_name,
                input_field_name="display_name",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=maybe_assertion_entity.description
                if maybe_assertion_entity
                else None,
            ),
            enabled=_merge_field(
                input_field_value=enabled,
                input_field_name="enabled",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=existing_assertion.mode == AssertionMode.ACTIVE
                if existing_assertion
                else None,
            ),
            schedule=_merge_field(
                input_field_value=schedule,
                input_field_name="schedule",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=existing_assertion.schedule
                if existing_assertion
                else None,
            ),
            detection_mechanism=_merge_field(
                input_field_value=detection_mechanism,
                input_field_name="detection_mechanism",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=SmartColumnMetricAssertion._get_detection_mechanism(
                    maybe_assertion_entity, maybe_monitor_entity, default=None
                )
                if maybe_assertion_entity and maybe_monitor_entity
                else None,
            ),
            sensitivity=_merge_field(
                input_field_value=sensitivity,
                input_field_name="sensitivity",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=maybe_monitor_entity.sensitivity
                if maybe_monitor_entity
                else None,
            ),
            exclusion_windows=_merge_field(
                input_field_value=exclusion_windows,
                input_field_name="exclusion_windows",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=maybe_monitor_entity.exclusion_windows
                if maybe_monitor_entity
                else None,
            ),
            training_data_lookback_days=_merge_field(
                input_field_value=training_data_lookback_days,
                input_field_name="training_data_lookback_days",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=maybe_monitor_entity.training_data_lookback_days
                if maybe_monitor_entity
                else None,
            ),
            incident_behavior=_merge_field(
                input_field_value=incident_behavior,
                input_field_name="incident_behavior",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=SmartColumnMetricAssertion._get_incident_behavior(
                    maybe_assertion_entity
                )
                if maybe_assertion_entity
                else None,
            ),
            tags=_merge_field(
                input_field_value=tags,
                input_field_name="tags",
                validated_assertion_input=assertion_input,
                validated_existing_assertion=existing_assertion,
                existing_entity_value=maybe_assertion_entity.tags
                if maybe_assertion_entity
                else None,
            ),
            created_by=existing_assertion.created_by
            or DEFAULT_CREATED_BY,  # Override with the existing assertion's created_by or the default created_by if not set
            created_at=existing_assertion.created_at
            or now_utc,  # Override with the existing assertion's created_at or now if not set
            updated_by=assertion_input.updated_by,  # Override with the input's updated_by
            updated_at=assertion_input.updated_at,  # Override with the input's updated_at (now)
        )

        return merged_assertion_input



import datetime
import logging
import os
import pathlib
import tempfile
import time
from enum import Enum
from typing import Any, Dict, Iterable, List, Literal, Optional, Tuple, Union, cast

import boto3
import duckdb
import pandas
import pyarrow as pa
import pyarrow.parquet as pq
from pydantic import BaseModel, field_validator

from acryl_datahub_cloud.elasticsearch.graph_service import BaseModelRow, SchemaField
from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import make_dataset_urn
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.metadata.schema_classes import (
    AuditStampClass,
    BooleanTypeClass,
    DatasetFieldProfileClass,
    DatasetProfileClass,
    DatasetPropertiesClass,
    NumberTypeClass,
    OperationClass,
    OperationTypeClass,
    OtherSchemaClass,
    SchemaFieldClass,
    SchemaFieldDataTypeClass,
    SchemaMetadataClass,
    StatusClass,
    StringTypeClass,
    TimeStampClass,
    TimeTypeClass,
)

logger = logging.getLogger(__name__)


class PartitioningStrategy(str, Enum):
    DATE = "date"
    SNAPSHOT = "snapshot"


class DatasetRegistrationSpec(BaseModel):
    """
    Specifications for dataset to register
    """

    soft_deleted: bool = (
        True  # When set to True, the dataset will be registered as a soft-deleted asset
    )
    profile: bool = (
        True  # When set to True, the dataset will be profiled before registering
    )
    last_modified: bool = (
        True  # When set to True, the last modified timestamp will be updated
    )
    operation_aspect: bool = (
        True  # When set to True, the operation aspect will be added to the MCE
    )


class FileStoreBackedDatasetConfig(ConfigModel):
    dataset_name: str
    dataset_urn: Optional[str] = (
        None  # If not set, it will be generated from the dataset_name
    )
    bucket_prefix: str
    store_platform: str = "s3"
    file_name: str = "data"
    file_extension: str = "parquet"
    file_compression: Literal[
        "gzip", "bz2", "brotli", "lz4", "zstd", "snappy", "none"
    ] = "snappy"
    file_overwrite_existing: bool = True
    snapshot_partitioning_strategy: str = PartitioningStrategy.DATE
    generate_presigned_url: bool = True
    presigned_url_expiry_days: int = 7
    dataset_registration_spec: DatasetRegistrationSpec = DatasetRegistrationSpec()
    file: Optional[str] = (
        None  # This is the file to be registered. When set, the file will be registered as a dataset immediately
    )

    datahub_platform: str = "acryl"

    @field_validator("snapshot_partitioning_strategy")
    @classmethod
    def validate_partitioning_strategy(cls, v):
        if v not in PartitioningStrategy._value2member_map_:
            raise ValueError(f"Unsupported partitioning strategy: {v}")
        return v

    @staticmethod
    def dummy():
        return FileStoreBackedDatasetConfig(dataset_name="none", bucket_prefix="none")


class DatasetMetadata(BaseModel):
    schemaFields: Optional[List[SchemaField]] = None
    displayName: Optional[str] = None
    description: Optional[str] = None


class DataHubBasedS3Dataset:
    def __init__(
        self,
        config: FileStoreBackedDatasetConfig,
        dataset_metadata: DatasetMetadata,
    ):
        self.s3_client = boto3.client("s3")
        self.config = config
        self.dataset_metadata = dataset_metadata
        self.opened_files: List[str] = []
        self.first_row = True
        self.current_record_batch: Optional[
            List[Union[Dict[str, Any], BaseModelRow]]
        ] = None
        self.local_file_path: str = (
            config.file if config.file else self._initialize_local_file()
        )
        self.file_writer: Optional[pq.ParquetWriter] = None
        self.schema = (
            pa.schema(
                [
                    pa.field(x.name, BaseModelRow.string_to_pyarrow_type(x.type))
                    for x in self.dataset_metadata.schemaFields
                ]
            )
            if self.dataset_metadata.schemaFields
            else None
        )

    def get_dataset_urn(self) -> str:
        return self.config.dataset_urn or make_dataset_urn(
            self.config.datahub_platform, self.config.dataset_name
        )

    def get_file_uri(self) -> str:
        return self.get_remote_file_uri(
            self.config.bucket_prefix,
            date=(
                datetime.date.today()
                if self.config.snapshot_partitioning_strategy
                == PartitioningStrategy.DATE
                else None
            ),
        )

    def _initialize_local_file(self) -> str:
        # create a temp dir and a file in the temp dir
        temp_path = tempfile.mkdtemp()
        bucket, key = self.get_file_uri().replace("s3://", "").split("/", 1)
        file_path = f"{temp_path}/{bucket}/{key}"
        pathlib.Path(file_path).parent.mkdir(parents=True, exist_ok=True)
        return file_path

    def append(self, row: Union[Dict[str, Any], BaseModelRow]) -> None:
        if self.first_row:
            self.first_row = False
            self.current_record_batch = []
            self.stringify_row: bool = False

            if not self.schema:
                if isinstance(row, BaseModelRow):
                    # BaseModelRow allows us to introspect the schema and
                    # generate an arrow schema from it
                    self.schema = row.arrow_schema()
                else:
                    # hail mary: infer schema from the first row and cast everything to string
                    self.schema = pa.schema([pa.field(key, pa.string()) for key in row])
                    self.stringify_row = True

            self._initialize_local_file()
            # Map compression names to PyArrow format (most are direct mappings)
            compression_map = {
                "gzip": "gzip",
                "bz2": "brotli",  # PyArrow doesn't support bz2, use brotli
                "brotli": "brotli",
                "lz4": "lz4",
                "zstd": "zstd",
                "snappy": "snappy",
                "none": "none",
            }
            compression = cast(
                Literal["gzip", "bz2", "brotli", "lz4", "zstd", "snappy", "none"],
                compression_map.get(self.config.file_compression, "snappy"),
            )
            self.file_writer = pq.ParquetWriter(
                self.local_file_path,
                self.schema,
                compression=compression,
            )
        if isinstance(row, (BaseModel, BaseModelRow)):
            # for anything extending BaseModel, we want to use the dict representation
            write_row: Dict[str, Any] = row.model_dump()
        elif isinstance(row, dict):
            write_row = row
        else:
            raise ValueError(
                f"Unsupported type for row: {type(row)}. Must be a dict or a BaseModel"
            )
        # At this point we know that write_row is a dict
        assert isinstance(write_row, dict)
        if self.stringify_row:
            write_row = {k: str(v) for k, v in write_row.items()}
        assert self.current_record_batch is not None
        self.current_record_batch.append(write_row)
        if len(self.current_record_batch) > 1024 * 1024:
            self._write_current_batch()

    def _write_current_batch(self):
        assert self.file_writer is not None
        assert self.schema is not None
        self.file_writer.write_batch(
            pa.RecordBatch.from_pandas(
                pandas.DataFrame(self.current_record_batch), schema=self.schema
            )
        )
        self.current_record_batch = []

    def commit(self) -> Iterable[MetadataChangeProposalWrapper]:
        if self.file_writer is not None and self.file_writer.is_open:
            self._write_current_batch()
            self.file_writer.close()
        if self.local_file_path:
            self.opened_files.append(self.local_file_path)
        if not os.path.exists(self.local_file_path):
            logger.info(f"File not found: {self.local_file_path}")
            return
        if os.path.getsize(self.local_file_path) == 0:
            logger.info(f"File is empty: {self.local_file_path}")
            return
        if self.config.store_platform == "s3":
            self.upload_file_to_s3()
        yield from self._register_dataset()

    def upload_file_to_s3(self):
        bucket, key = self.get_file_uri().replace("s3://", "").split("/", 1)
        assert self.s3_client is not None
        assert self.local_file_path is not None
        logger.info(f"Uploading {self.local_file_path} to s3://{bucket}/{key}")
        self.s3_client.upload_file(self.local_file_path, bucket, key)

    def _register_dataset(self) -> Iterable[MetadataChangeProposalWrapper]:
        dataset_urn = self.get_dataset_urn()
        assert self.local_file_path is not None
        yield from self.register_dataset(
            dataset_urn, self.get_file_uri(), self.local_file_path
        )

    def _generate_presigned_url(self, s3_uri: str) -> str:
        """
        Generate a pre-signed URL for downloading an object from S3.

        Args:
        - s3_uri (str): The S3 URI of the object to be downloaded (e.g., 's3://bucket-name/object-key').
        - expiration (int): Expiration time for the pre-signed URL in seconds (default is 3600 seconds).

        Returns:
        - str: The pre-signed URL for downloading the object.
        """
        # Parse the S3 URI to get bucket name and object key
        bucket_name, object_key = s3_uri.replace("s3://", "").split("/", 1)
        expiration_seconds = self.config.presigned_url_expiry_days * 24 * 60 * 60
        # Generate a pre-signed URL for downloading the object
        presigned_url = self.s3_client.generate_presigned_url(
            "get_object",
            Params={"Bucket": bucket_name, "Key": object_key},
            ExpiresIn=expiration_seconds,
        )

        return presigned_url

    def get_remote_file_uri(
        self, dataset_uri_prefix: str, date: Optional[datetime.date] = None
    ) -> str:
        if self.config.snapshot_partitioning_strategy == PartitioningStrategy.DATE:
            assert date is not None
            assert dataset_uri_prefix is not None
            # TODO: This should be date.strftime('year=%Y/month=%m/day=%d') for correct time partition in S3
            return f"{dataset_uri_prefix.rstrip('/')}/{date.strftime('%Y-%m-%d')}/{self.config.file_name}.{self.config.file_extension}"
        elif (
            self.config.snapshot_partitioning_strategy == PartitioningStrategy.SNAPSHOT
        ):
            return f"{dataset_uri_prefix.rstrip('/')}/{self.config.file_name}.{self.config.file_extension}"
        else:
            raise ValueError(
                f"Unsupported partitioning strategy: {self.config.snapshot_partitioning_strategy}"
            )

    def _generate_schema_metadata(
        self, duckdb_columns: List[Tuple[str, Any]]
    ) -> SchemaMetadataClass:
        def get_type_from_dtype(dtype: str) -> SchemaFieldDataTypeClass:
            if "int" in dtype or "float" in dtype or "number" in dtype:
                return SchemaFieldDataTypeClass(type=NumberTypeClass())
            elif "bool" in dtype:
                return SchemaFieldDataTypeClass(type=BooleanTypeClass())
            elif "datetime" in dtype or "date" in dtype:
                return SchemaFieldDataTypeClass(type=TimeTypeClass())
            else:
                return SchemaFieldDataTypeClass(type=StringTypeClass())

        # for each column -> generate min, max, avg, distinct count, null count
        # for the dataset -> generate row count, number of columns
        # for each column -> generate data type
        audit_stamp = AuditStampClass(
            time=int(time.time()) * 1000, actor="urn:li:corpuser:datahub"
        )
        schema_metadata = SchemaMetadataClass(
            created=audit_stamp,
            lastModified=audit_stamp,
            hash="",
            platform="urn:li:dataPlatform:s3",
            version=0,
            primaryKeys=[],
            schemaName="",
            fields=[],
            platformSchema=OtherSchemaClass(rawSchema=str(self.schema)),
        )
        for column in duckdb_columns:
            # generate data type
            data_type = str(column[1]).lower()
            schema_metadata.fields.append(
                SchemaFieldClass(
                    fieldPath=column[0],
                    type=get_type_from_dtype(data_type),
                    nativeDataType=str(data_type),
                )
            )
        return schema_metadata

    def _generate_dataset_profile_and_schema(
        self, file: str
    ) -> Tuple[DatasetProfileClass, SchemaMetadataClass]:
        # for each column -> generate min, max, avg, distinct count, null count
        # for the dataset -> generate row count, number of columns
        # for each column -> generate data type

        from_fragment = f"FROM parquet_scan('{file}')"
        with duckdb.connect() as conn:
            # duckdb.execute(
            #     f"CREATE TABLE forms AS SELECT * FROM parquet_scan('{file}')"
            # )
            row_count = conn.execute("SELECT COUNT(*) " + from_fragment).fetchall()[0][
                0
            ]
            columns = conn.execute("SELECT * " + from_fragment + " LIMIT 1").description
            logger.info("Generating dataset profile")
            assert columns is not None
            dataset_profiles = DatasetProfileClass(
                timestampMillis=int(time.time()) * 1000,
                rowCount=row_count,
                columnCount=len(columns),
                fieldProfiles=[],
            )
            assert isinstance(columns, list)
            for column in columns:
                # generate min, max, avg, distinct count, null count
                column_name = column[0]
                logger.info(f"Generating field profile for {column_name}")
                data_type = str(column[1]).lower()
                if "int" in data_type or "float" in data_type:
                    query = (
                        f"SELECT COUNT(DISTINCT {column_name}), COUNT(*) - COUNT({column_name}), MIN({column_name}), MAX({column_name}), AVG({column_name})"
                        + from_fragment
                    )
                    (
                        distinct_count,
                        null_count,
                        min_value,
                        max_value,
                        avg_value,
                    ) = conn.execute(query).fetchall()[0]
                elif "date" in data_type:
                    query = (
                        f"SELECT COUNT(DISTINCT {column_name}), COUNT(*) - COUNT({column_name}), MIN({column_name}), MAX({column_name})"
                        + from_fragment
                    )
                    (distinct_count, null_count, min_value, max_value) = conn.execute(
                        query
                    ).fetchall()[0]
                    avg_value = None
                else:
                    query = (
                        f"SELECT COUNT(DISTINCT {column_name}), COUNT(*) - COUNT({column_name})"
                        + from_fragment
                    )
                    (distinct_count, null_count) = conn.execute(query).fetchall()[0]
                    min_value = None
                    max_value = None
                    avg_value = None
                # get 10 sample values
                sample_values = [
                    str(x[0])
                    for x in conn.execute(
                        f"SELECT {column_name} " + from_fragment + " using SAMPLE 10"
                    ).fetchall()
                ]
                print(
                    f"Column: {column}, Min: {min_value}, Max: {max_value}, Avg: {avg_value}, Distinct: {distinct_count}, Null: {null_count}, Data Type: {data_type}"
                )
                field_profile = DatasetFieldProfileClass(
                    fieldPath=column[0],
                    min=str(min_value),
                    max=str(max_value),
                    mean=str(avg_value),
                    uniqueCount=distinct_count,
                    uniqueProportion=float(distinct_count) * 100.0 / float(row_count),
                    nullCount=null_count,
                    nullProportion=float(null_count) * 100.0 / float(row_count),
                    sampleValues=sample_values,
                )
                assert dataset_profiles.fieldProfiles is not None
                dataset_profiles.fieldProfiles.append(field_profile)
            logger.info("Generated dataset profile")
            schema_metadata = self._generate_schema_metadata(
                [(col[0], col[1]) for col in columns]
            )
        return dataset_profiles, schema_metadata

    def register_dataset(
        self,
        dataset_urn: str,
        physical_uri: str,
        local_file: str,
    ) -> Iterable[MetadataChangeProposalWrapper]:
        aspects: List = []
        mcps: List[MetadataChangeProposalWrapper] = self._update_presigned_url(
            dataset_urn, physical_uri
        )
        current_time_millis = int(time.time()) * 1000
        aspects.extend(
            [
                StatusClass(
                    removed=self.config.dataset_registration_spec.soft_deleted  # Registering this is as a visible asset as we have search access policies to ensure only admins can see it
                ),
                OperationClass(
                    timestampMillis=current_time_millis,
                    operationType=OperationTypeClass.UPDATE,
                    lastUpdatedTimestamp=current_time_millis,
                ),
            ]
        )

        if self.config.dataset_registration_spec.profile:
            (
                dataset_profiles,
                schema_metadata,
            ) = self._generate_dataset_profile_and_schema(local_file or physical_uri)
            aspects.append(dataset_profiles)
            aspects.append(schema_metadata)
        mcps.extend(
            MetadataChangeProposalWrapper.construct_many(
                entityUrn=dataset_urn,
                aspects=aspects,
            )
        )
        yield from mcps

    def update_presigned_url(
        self, dataset_properties: Optional[DatasetPropertiesClass]
    ) -> List[MetadataChangeProposalWrapper]:
        return self._update_presigned_url(
            self.get_dataset_urn(), self.get_file_uri(), dataset_properties
        )

    def _update_presigned_url(
        self,
        dataset_urn: str,
        physical_uri: str,
        dataset_properties: Optional[DatasetPropertiesClass] = None,
    ) -> List[MetadataChangeProposalWrapper]:
        if self.config.generate_presigned_url:
            external_url = self._generate_presigned_url(physical_uri)
        else:
            external_url = physical_uri

        logger.info(
            f"Registering dataset {dataset_urn} with physical URI {physical_uri}"
        )
        if dataset_properties is not None:
            dataset_properties.externalUrl = external_url
        else:
            current_time_millis = int(time.time()) * 1000
            dataset_properties = DatasetPropertiesClass(
                name=self.dataset_metadata.displayName or self.config.dataset_name,
                description=self.dataset_metadata.description or "",
                externalUrl=external_url,
                uri=None,  # Unfortunately, there is a bug in URI validation that requires this to be None
                lastModified=TimeStampClass(
                    time=current_time_millis, actor="urn:li:corpuser:datahub"
                ),
                customProperties={
                    "physical_uri": physical_uri,
                },
            )

        return MetadataChangeProposalWrapper.construct_many(
            entityUrn=dataset_urn,
            aspects=[dataset_properties],
        )
